package sorter;

import rac.*;

/**
 * A Template Method design pattern abstract sorter.  
 * An abstract ordering strategy is required that must be matched to the type of objects being sorted.
 */
public abstract class ASorter {
  private LRSStackFactory fact = new LRSStackFactory();

  protected AOrder aOrder;

  /**
   * The constructor for this class.
   * @param aOrder The abstract ordering strategy to be used by any subclass.
   */
  protected ASorter(AOrder aOrder) {
    this.aOrder = aOrder;
  }

  /**
   * 
   */
  private void sortCommand(final Object[] A, final int lo, final int hi, final IRAContainer stack) {
    if (lo < hi) {
      final int s = split(A, lo, hi);
      
      pushJoin(A, lo, s, hi, stack);
//      stack.put(new ILambda() {
//        public Object apply(Object... nu) {
//          join(A, lo, s, hi);
//          return null;
//        }
//      });
      
      /**
       * Compare the size of the upper and lower subarrays.  Then,
       * push the command to sort the larger subarray first.  Thus,
       * the command to sort the smaller subarray is on the top of
       * the command stack.  By sorting the smaller subarray before
       * the larger subarray, we ensure that the command stack is
       * never more than O(log n) deep, where n is the size of the
       * array.
       */
      if ((hi + 1 - s > s - lo)) {
        /**
         * The upper array is larger.
         */
        stack.put(new ILambda() {
          public Object apply(Object... nu) {
            sortCommand(A, s, hi, stack);
            return null;
          }
        });
        
        stack.put(new ILambda() {
          public Object apply(Object... nu) {
            sortCommand(A, lo, s - 1, stack);
            return null;
          }
        });
      } else {
        /**
         * The lower array is larger.
         */ 
        stack.put(new ILambda() {
          public Object apply(Object... nu) {
            sortCommand(A, lo, s - 1, stack);
            return null;
          }
        });
        
        stack.put(new ILambda() {
          public Object apply(Object... nu) {
            sortCommand(A, s, hi, stack);
            return null;
          }
        });
      }
    }
  }
  
  /**
   * Sorts by doing a split-sort-sort-join.  Splits the original array into two subarrays,
   * recursively sorts the split subarrays, then re-joins the sorted subarrays together.
   * 
   * The recursion is implemented through an explicit command stack instead of method
   * calls.  The commands are "sort a subarray" and "join two sorted subarrays".
   * 
   * @param A the array A[lo:hi] to be sorted.
   * @param lo the low index of A.
   * @param hi the high index of A.
   */
  public final void sort(final Object[] A, final int lo, final int hi) {
    final IRAContainer stack = fact.makeRAC();
    int maxDepth = 0;
    
    /**
     * Push the first sort command.
     */ 
    stack.put(new ILambda() {
      public Object apply(Object... nu) {
        sortCommand(A, lo, hi, stack);
        return null;
      }
    });

    /**
     * Execute the command at the top of the command stack until the stack is empty.
     */ 
    while (!stack.isEmpty()) {
      if (maxDepth < stack.count())
        maxDepth = stack.count();
      ILambda command = (ILambda)stack.get();
      command.apply();
    }
    System.out.println("Maximum stack depth = " + maxDepth);
  }

  /**
   * Splits A[lo:hi] into A[lo:s-1] and A[s:hi] where s is the returned value of this function.
   * @param A the array A[lo:hi] to be sorted.
   * @param lo the low index of A.
   * @param hi the high index of A.
   */
  protected abstract int split(Object[] A, int lo, int hi);

  /**
   * Joins sorted A[lo:s-1] and sorted A[s:hi] into A[lo:hi].
   * @param A A[lo:s-1] and A[s:hi] are sorted.
   * @param lo the low index of A.
   * @param hi the high index of A.
   */
  protected abstract void join(Object[] A, int lo, int s, int hi);

  /**
   * 
   */
  protected abstract void pushJoin(Object[] A, int lo, int s, int hi, IRAContainer stack);
  
  /**
   * An accessor method for the abstract ordering strategy.
   * @param aOrder 
   */
  public void setOrder(AOrder aOrder) {
    this.aOrder = aOrder;
  }
}

